/*+++++++++++   
.LANGUAGE       C
.IDENTIFICATION coopix.c
.AUTHOR         Francois Ochsenbein [ESO-IPG]
.KEYWORDS       C Software data-base & Documentation
.ENVIRONMENT    
.COMMENTS       Convert projection into pixels and vice-versa
 
.VERSION 1.0	07-jan-2003
.VERSION 1.1	25-Sep-2010: -version option

-------------------------------------------------------------------------*/

#define VERSION "1.1"

#include <astropos.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <tranum.h>			/* Interpret text */

double atof();	

static char *projections[] = {		/* List of projections */
    "tan",       (char *)_PROJ_TAN_,
    "tan2",      (char *)_PROJ_TAN2_,
    "stg",       (char *)_PROJ_STG_,
    "sin",       (char *)_PROJ_SIN_,
    "sin2",      (char *)_PROJ_SIN2_,
    "zea",       (char *)_PROJ_ZEA_,
    "arc",       (char *)_PROJ_ARC_,
    "schmidt",   (char *)_PROJ_SCHMIDT_,
    "aitoff",    (char *)_PROJ_AITOFF_,
    "ait",       (char *)_PROJ_AIT_,
    "gls",       (char *)_PROJ_GLS_,
    "mercator",  (char *)_PROJ_MERCATOR_,
    "mer",       (char *)_PROJ_MER_,
    "lam",       (char *)_PROJ_LAM_,
    "lambert",   (char *)_PROJ_LAMBERT_,
    "tsc",       (char *)_PROJ_TSC_,
    "qsc",       (char *)_PROJ_QSC_,
    (char *)0
} ;

static char usage[] = "\
Usage: coopix [-1] [-fxy] -lims x0,y0:x1,y1  [-c center] [-r radius] [-x] \n\
       -proj pname";
static char help[] = "\
   Convert coordinate into pixel, or vice-versa\n\
    -1: compute position from pixel (default is pexl from position)\n\
  -fxy: when one axis is flipped (-fx for x-axis, -fy for y-axis, -fxy both)\n\
  -lim: define the limiting pixels, default -1-1:+1+1\n\
    -c: center of the projection (defaut=0+0)\n\
    -p: projection type\n\
    -r: radius (default 180)\n\
";

static void list_projections()
/*++++++++++++++++
.PURPOSE  list the projections
.RETURNS  
.REMARKS  
-----------------*/
{
  char **a;
    for (a=projections; *a; a+=2) printf("%-10s", *a);
    putchar('\n');
}

static char *alook2(char **table, char *col1)
/*++++++++++++++++
.PURPOSE  Find the correspondance of col2 in a 2-col table
.RETURNS  The 2nd column / NULL
.REMARKS  Linear search
-----------------*/
{
  char **a ;
    if (table) for (a=table; *a; a += 2) {
        if (strcmp(*a, col1) == 0 ) return(a[1]) ;
    }
    return ((char *)0) ;
}

static double max_y(int proj, double radius)
/*++++++++++++++++
.PURPOSE  Give the maximal y / 0 if error
.RETURNS  Max value of y / 0
-----------------*/
{
  double o[2], p[2];

    if (radius >= 90.) switch(proj) {
      case _PROJ_TAN_:	
      case _PROJ_SIN_:	
	return(0.);
      case _PROJ_LAMBERT_:
      case _PROJ_TSC_:
      case _PROJ_QSC_:
      case _PROJ_GLS_:
      case _PROJ_AITOFF_: 
	radius = 90.; break;
    }
    o[0] = 0; o[1] = radius;
    if (!tr_op(proj, o, p)) return(0.);
    return(p[1]);
}

int main(int argc, char **argv)
{
  static char buf[128];
  static double p0[] = { -1, -1 };
  static double p1[] = {  1,  1 };
  static double oc[] = {  0,  0 };
  static double pc[] = {  0,  0 };
  static double factor = 1;
  double f, u0[3], R0[3][3], o[2], p[2], u[3];
  static int flip= 0;	/* 1=x, 2=y */
  static int inv = 0;
  static int rot = 0;
  static int status = 0;
  static char fmt[32] = "%08.5f%+08.5f";
  static int proj = 0;
  static double radius = 180;
  char *b;
  int nd;

    while (--argc > 0) {
    	b = *++argv;
	if (b[0] == '-') switch(b[1]) {
	  case 'h':	/* Help */
	    printf("%s%s", usage, help);
	    list_projections();
	    exit(0);
	  case 'v':	/* Assume -version */
	    printf("coopix (CDS) Version %s\n", VERSION);
	    exit(0);
	  case '1':	/* Reverse */
	    inv = 1;
	    continue;
	  case 'f':	/* Flipped */
	    for (b += 2; *b; b++) switch(*b) { 
	      case 'x': flip |= 1; continue;
	      case 'y': flip |= 2; continue;
	      default:  
		fprintf(stderr, "++++Ignored '%c' in argument: %s\n", 
		  *b, *argv);
	    }
	    continue;
	  case 'c':	/* Center */
	    --argc; ++argv;
	    if (!*argv) goto bad_argument;
	    if (tr_ao(buf, strlen(buf), oc)<0) {
	        fprintf(stderr, "****Invalid center (-c argument)\n%s\n",
		  usage);
		exit(1);
	    }
  	    tr_ou(oc,u0);
  	    tr_oR(oc,R0);
	    rot = 1;
	    continue;
	  case 'p':	/* Projection type */
	    --argc; ++argv;
	    if (!*argv) b = "";
	    else proj  = (int)((intp)alook2(projections, b = *argv));
	    if (!proj) {
	        fprintf(stderr, "****Invalid projection '%s'\n", b);
		list_projections();
		exit(1);
	    }
	    continue;
	  case 'r':	/* Radius */
	    --argc; ++argv;
	    if (!*argv) goto bad_argument;
	    if (tr_af(buf, strlen(buf), &radius)<0) {
	        fprintf(stderr, "****Invalid radius (-r argument)\n%s\n",
		  usage);
		exit(1);
	    }
	  case 'l':	/* Limits */
	    --argc; ++argv;
	    if (!*argv) goto bad_argument;
	    b = strchr(*argv, ':');
	    if (!b) goto bad_limits;
	    if (tr_af1(*argv, b-*argv, p0, 2) != 2) goto bad_limits;
	    ++b;
	    if (tr_af1(b, strlen(b), p1, 2) != 2) goto bad_limits;
	    pc[0] = (p0[0] + p1[0])/2.;
	    pc[1] = (p0[1] + p1[1])/2.;
	    factor = (p1[1] - p0[1])/2.;
	    nd = 5;
	    if (factor <= 0.1) nd=6;
	    for (f=factor; (f>=5) && (nd>0); f/=10, nd--) ;
	    sprintf(fmt, "%%0%d.%df%%+0%d.%df", 8, nd, 8, nd);
	    continue;
	}

      bad_limits:
	fprintf(stderr, "****Invalid limits: %s\n%s\n", *argv, usage);
	exit(1);
      bad_argument:
	fprintf(stderr, "****Invalid argument: %s\n%s\n", *argv, usage);
    	exit(1);
    }
    if (!proj) {
	fprintf(stderr, "****Missing Projection Type\n%s\n", usage);
	exit(1);
    }
    f = max_y(proj, radius) ;
    factor /= f;

    /* Get now the positions, and compute */

    while(1) {
      	if (isatty(0)) printf("----Give %s position: ", inv ? "pixel" : "sky");
	if (!fgets(buf, sizeof(buf), stdin)) break;
	for (b=buf+strlen(buf)-1; (b>=buf) && isspace(*b); --b);
	*++b = 0;
	if ((buf[0] == '#') || (buf[0] == 0)) {	/* Comment */
	    printf("%s\n", buf);
	    continue;
	}

	if (inv) {
	    if (tr_af1(buf, (b-buf), p, 2) != 2) {
		printf("#+++Ignored record: %s\n", buf);
		status = 2;
		continue;
	    }
	    p[0] = (p[0] - pc[0])/factor;
	    p[1] = (p[1] - pc[1])/factor;
	    if (flip&1) p[0] = -p[0];
	    if (flip&2) p[1] = -p[1];
	    if (!tr_po(proj, p, o)) {
	        printf("#+++Outside limits: %s\n", buf);
		continue;
	    }
	    if (rot) {
	        tr_pu(proj, p, u);
		tr_uu1(u, u, R0);
		tr_uo(u, o);
	    }
	    printf("%08.4f%+08.4f\n", o[0], o[1]);
	    continue;
	}

	if (tr_ao(buf, (b-buf), o) <= 0) {
	    printf("#+++Ignored record: %s\n", buf);
	    status = 2;
	    continue;
	}
	if (rot) {
	    tr_ou(o, u);
	    tr_uu(u, u, R0);
	    tr_up(proj, u, p);
	}
	else tr_op(proj, o, p);
	if (flip&1) p[0] = -p[0];
	if (flip&2) p[1] = -p[1];
	p[0] = factor*p[0] + pc[0];
	p[1] = factor*p[1] + pc[1];
	printf(fmt, p[0], p[1]);
	putchar('\n');
    }
  
    exit(status);
}  
