/*+++++++++++++++++++++++++++++
.TYPE		Module
.IDENTIFICATION dist2.c
.AUTHOR  	Francois Ochsenbein [ESO]  
.KEYWORDS       Astronomical Coordinate Transformations
.VERSION 1.0    19-Jul-1994
.VERSION 1.1    30-May-1996: Test program
.COMMENTS
\begin{TeX}

The routines provided in this module compute the relative position
\quad $(\rho, \theta)$ 
between 2 points on the celestial sphere.
The positions of the points can be specified
as angles (in degrees, {\em dist2\_o}), 
vectors (direction cosines, {\em dist2\_u}), 
or gnomonic projections ({\em dist2\_p}).

\end{TeX}

---------------------------------*/
 
#include <astropos.h>

/*========================================================================*/
int dist2_o(q0, q2, r)
/*+++                   
.PURPOSE Computes (rho,theta) from equatorial positions
.RETURNS 0 if distance is zero pr 180deg (no angle) / 1
.METHOD  Direct computation
---*/
  double q0[2];	/* IN: position (ra,dec) of the first point */
  double q2[2];	/* IN: position (ra,dec) of the second point*/
  double r[2];	/* OUT: r[0] = distance (deg), r[1] = pos.angle (deg) */
{ 
  double R[3][3], u[3];
  int i;

    tr_oR(q0, R);	/* Compute Rotation Matrix, R[0] = u0 */
    tr_ou(q2, u);
    tr_uu(u,u,R);	/* Components of q2 in local frame of q0 */

    r[0] = u[1]*u[1] + u[2]*u[2];	/* Tan(distance) */

    if (r[0]) i = 1, r[1] = atan2d(u[1], u[2]), r[0] = atand(sqrt(r[0])/u[0]);
    else      i = 0, r[1] = 0;
    if (r[0] < 0) r[0] += 180.;
    if (r[1] < 0) r[1] += 360.;
    return ( i );
}

/*========================================================================*/

int dist2_u(u0, u2, r)
/*+++
.PURPOSE Computes (rho,theta) from direction cosines
.RETURNS 0 if distance is zero (no angle) / 1
---*/
  double u0[3];	/* IN: : dir. cosines (x,y,z) of the first point */
  double u2[3];	/* IN: : dir. cosines (x,y,z) of the second point*/
  double r[2];	/* OUT: r[0] = distance (deg), r[1] = pos.angle (deg) */
{
  double R[3][3], u[3];
  int i;

    tr_uR(u0, R);	/* Compute Rotation Matrix, R[0] = u0 */
    tr_uu(u2,u,R);	/* Components of u2 in local frame of u0 */

    r[0] = u[1]*u[1] + u[2]*u[2];	/* tan(distance) */

    if (r[0]) i = 1, r[1] = atan2d(u[1], u[2]), r[0] = atand(sqrt(r[0])/u[0]);
    else      i = 0, r[1] = 0;
    if (r[0] < 0) r[0] += 180.;
    if (r[1] < 0) r[1] += 360.;
    return ( i );
}

/*========================================================================*
		Test Program
 *========================================================================*/
#ifdef TEST
main()
/*++++++++++++++++
.PURPOSE  Test program...
.RETURNS  to OS
-----------------*/
{
  char buf[512]; 
  double o0[2], o1[2], r[2];

    while(1) {
    	printf("Give central angle: "); 
	if (! gets(buf)) break;
	if (tr_aq(buf, strlen(buf), o0) <= 0) continue;
	printf("================== %10.6f%+10.6f\n", o0[0], o0[1]);
	while(1) {
	    printf("      Other  angle: ");
	    if (! gets(buf)) break;
	    if (tr_aq(buf, strlen(buf), o1) <= 0) continue;
	    printf("      ------------ %10.6f%+10.6f\n", o1[0], o1[1]);
	    printf("Dist      %10.2f\"\n", dist_o(o0, o1)*3600.);
	    printf("Dist2 (%d) ", dist2_o(o0, o1, r));
	    printf("%10.2f\", %5.1fdeg\n", r[0]*3600., r[1]);
	}
    }
}
#endif
