/*++++++++++++++++++++++++
.TYPE		Module
.IDENTIFICATION	qbshow.c
.AUTHOR		Francois Ochsenbein [ESO]  
.VERSION  1.0	06-Jan-1999: Main Program
.VERSION  1.1	25-Oct-2007
.VERSION  1.2	04-Sep-2008: -a option.
.Language	C
.ENVIRONMENT	Any
.KEYWORDS	Show Qbox Properties
.COMMENTS	See qbox.c
--------------------*/ 

#define VERSION "1.2"

#include <qbox.h>		/* Definitions for qboxes	*/
#include <trigo.h>		/* Spherical Function		*/
#include <math.h>		/* All math functions		*/
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#define issign(c)		(((c)=='-')||((c)=='+'))

static char opted ;		/* 0 = ascii, 1 = tex */
static char opta = 0;		/* -a option (all levels) */
static char optc = 0;
#define LEVEL 8			/* Defaut level */
static int  level = -1;
static char eds[] = "an2s";
 
static char usage[] = "\
Usage: qbshow [-a[..]] [-c] [-lev] [-tex] [-e[a|n|s]] {0x:n|n|position}..." ;
static char help[] = "\
This program shows the properties of the \"qboxes\":\n\
        the level, center, max radius, surface.\n\
  -a  : give values for all levels until the one specified\n\
  -a..: as -a, but display only first/last for each level\n\
  -c  : input as coordinates\n\
  -lev: specify the level (range 0..12)\n\
  -tex: edit as a LaTeX table\n\
 Cells may be written:\n\
   - as numeric ([1-6].[0-3]*)  or alphabetic ([N-S][a-p]*[0-3]?) qbox\n\
   - as a single integer number\n\
   - as 2 short integers when a colon exists (ex. 15:-1234)\n\
 Cells may be edited as alphabetic (-ea), numeric (-en) or 2 numbers (-e2)\n\
" ;

static char tex_init[] = "\
\\begin{center}\n\
\\begin{tabular}{|r|l|r|r|r|r|}\\hline\n\
\\multicolumn{1}{c}{{\\bf Lev.}} &\n\
\\multicolumn{1}{c}{{\\bf Qbox}} &\n\
\\multicolumn{1}{c}{{\\bf Center} ($\\d$)} &\n\
\\multicolumn{1}{c}{{\\bf Size}} &\n\
\\multicolumn{1}{c}{{\\bf Radius}} &\n\
\\multicolumn{1}{c}{{\\bf Surface} ($\\d^2$)} \\\\ \\hline\n\
" ;
static char asc_init[] = "\
-----------------------------------------------------------------------------\n\
Lev    Qbox           Center (deg)      Size    Radius       Surface\n\
-----------------------------------------------------------------------------\n\
" ;

static char tex_end[] = "\
\\hline\\end{tabular}\\end{center}\n" ;
static char asc_end[] = "\
-----------------------------------------------------------------------------\n\
" ;

static int check_num(char *param, int len)
/*+++++
.PURPOSE Check the number.
.RETURNS 0 (OK) / -1 (err)
-----*/
{
  char *p, *e ;
    if (len <= 0) len = strlen(param) ;
    p = param, e = p + len ;
    if (issign(*p)) p++ ;
    while ((p < e) && isdigit(*p)) p++ ;

    if (p == e) return(0) ;

    fprintf(stderr, "****Bad '") ;
    for (p=param; p<e; p++) fputc(*p, stderr) ;
    fprintf(stderr, "' number, ignored\n") ;
    return(-1) ;
}

static void pr_cell(int qbid, int recursive)
/*+++++
.PURPOSE Print the elements of a cell
.RETURNS  ---
-----*/
{
  static char init_done ;
  char buf[20], *p, ed;
  double o[2] ;
  int lev;
    
    ed = eds[(opted&7)-1];	/* Edition */
    if (!init_done) {
	if (opted&7) {		/* Remove a few blanks */
	  int rmb;		/* Blanks to remove... */
	    p = strstr(asc_init, "    Center");
	    rmb = ed == '2' ? 1 : 4;
	    strcpy(p, p+rmb);
	}
	printf("%s", opted&0x10 ? tex_init : asc_init) ;
	init_done = 1 ;
    }

    lev = qbox_level(qbid);
    if (lev<0) {
	if (qbid != 0) {	/* Bad qbox number */
	    fprintf(stderr, "#*** Bad qbox number: %08x = %d\n", qbid, qbid);
	    return;
	}
	/* Whole sky... */
	if (recursive) {
	    for (qbid=9; qbid<15; qbid++) {
		putchar('\n');	/* Blank line between 2 basic levels */
		pr_cell(qbid, recursive);
	    }
	    return;
	}
    }

    if (recursive) { int shift, m, qb, qm;
	for (shift=(lev<<1); shift>=0; shift -= 2)
	    pr_cell(qbid>>shift, 0);
	shift = 0;
	while (lev < level) {
	    lev++; shift += 2; 
	    qb = qbid<<shift; 
	    qm = qb|((1<<shift)-1);
	    pr_cell(qb, 0); 
	    if ((recursive&2) && (ed != '2')) ;
	    else while (++qb < qm) {
		if (recursive&2) {	/* Here only when ed=='2' */
		    m = qb&0x7fff; 
		    if ((m==0) || (m==0x7fff))
			pr_cell(qb, 0);
		}
		else pr_cell(qb, 0);
	    }
	    pr_cell(qm, 0);
	}
	return;
    }

    printf("%3d", qbox_level(qbid)) ;
    if (opted&0x10) putchar('&') ;
    printf("  ");
    /* Various editions, depending on option */
    p = buf;
    if (opted&7) switch(eds[(opted&7)-1]) {
      case 'a':	
	  printf("%-10s", qbox8a(qbid, 0));
	  break;
      case '2':
	  printf("%7d:%06d", qbid>>16, (qbid<<16)>>16);
	  break;
      case 'n': 
	  printf("%10d", qbid);
	  break;
    }
    else printf("%-14s", qbox2a(qbid)) ;
    qbox_co(qbid, o) ;
    if (opted&0x10) putchar('&') ;
    printf(" %08.4f%+08.4f", o[0], o[1]) ;

    if (opted&0x10) putchar('&') ;
    o[0] = qbox_size(qbid);
    if (o[0] <= 0) printf(" %8s", "--- ");
    else         printf(" %8.4f", o[0]);

    if (opted&0x10) putchar('&') ;
    printf(" %8.4f", qbox_radius(qbid)) ;

    if (opted&0x10) putchar('&') ;
    printf(" %14.8f", qbox_surf(qbid)) ;

    if (opted&0x10) printf("\\\\") ;
    putchar('\n') ;
}

static int tr_ao(char *text, double *o)
/*+++++
.PURPOSE Get 2 coordinates
.RETURN  Number of bytes interpreted
-----*/
{
  char *p;
    p = text;
    o[0] = o[1] = 0;
    if (issign(*p)) p++;
    while (isdigit(*p)) p++;
    if (*p == '.') for (++p; isdigit(*p); p++) ;
    o[0] = atof(text);
    while (*p == ' ') p++;
    o[1] = atof(p);
    if (issign(*p)) p++;
    while (isdigit(*p)) p++;
    if (*p == '.') for (++p; isdigit(*p); p++) ;
    return(p-text);
}

int main(int argc, char **argv)
/*+++++
.PURPOSE Test implementation
.RETURN To OS
-----*/
{
  char *ac, *p ;
  int  count = 0;
  double o[2];
  int i, h[2];

    while(--argc > 0) {
	p = *++argv ;
	if (*p == '-') switch(p[1]) {

	  case 'h':
	    printf("%s\n%s", usage, help) ;
	    exit(0) ;

	  case 'a':	/* All levels  */
	    opta = 1;
	    if (ispunct(p[2])) opta = 2;
	    continue;

	  case 'c':	/* Coordinates */
	    optc = 1;
	    continue ;

	  case 't':	/* -tex	*/
	    opted |= 0x10 ;
	    continue ;

	  case 'v':	/* Version */
	    printf("qbshow (CDS) Version %s\n", VERSION);
	    exit(0);

	  case 'e':
	    opted &= ~0x10;	/* Remove options */
	    ac = strchr(eds, p[2]);
	    if (ac) { 
		i = ac-eds; 
		if (i>2) i=2;
		opted |= (1+i);
	    }
	    else {
		fprintf(stderr, "****Can't understand edit option: %s\n", p);
		fprintf(stderr, "%s\n", usage) ;
		exit(1) ;
	    }
	    continue ;

	  case 'l':		/* Level */
	    for (ac=p+1; isalpha(*ac); ac++) ;
	    if (isdigit(*ac)) {
		level = atoi(ac);
		continue;
	    }
	    else {
		fprintf(stderr, "****Can't understand level option: %s\n", p);
		fprintf(stderr, "%s\n", usage) ;
		exit(1) ;
	    }
	    
	  default:		/* accept e.g. -8 = Level 8 */
	    if (isdigit(p[1])) {
		level = atoi(p+1);
		if (level > 12) { 
		    fprintf(stderr, 
			    "#+++Level (%d) limited to 12+++\n", level);
		    level = 12;
		}
		for (i=1; isdigit(p[i]); i++) ;
		if (p[i] == 'a') 
		    opta = ispunct(p[i+1]) ? 2 : 1 ;
		continue;
	    }
	    else break;
	    if (isdigit(p[1])) break;
	    fprintf(stderr, "****Can't understand argument: '%s'\n", p) ;
	    fprintf(stderr, "%s\n", usage) ;
	    exit(1) ;
	}
	count++;
	if (optc) {
	    if (level<0) level = LEVEL;		/* Default */
	    tr_ao(p, o);
	    pr_cell(qbox__o(level, o), opta);
	    continue;
	}
	if ((ac = strchr(p, ':'))){	/* 2 numbers */
	    if (check_num(p, ac-p)) continue ;
	    h[0] = atoi(p) & 0xffff;
	    p = ++ac ;
	    if (check_num(p, -1)) continue ;
	    h[1] = atoi(p) & 0xffff;
	    pr_cell((h[0]<<16)|h[1], opta) ;
	    continue ;
	}
	else if (isalpha(*p) || strchr(p, '.')) {
	    pr_cell(qbox_a(p), opta);
	    continue;
	}
	else {
	    if (check_num(p, -1)) continue ;
	    pr_cell(atoi(p), opta) ;
	    continue ;
	}
    }
    if (count == 0) 
	 printf("%s\n%s", usage, help) ;
    else printf("%s", opted&0x10 ? tex_end : asc_end) ;
    exit(0) ;
}
