/*++++++++++++++++++++++
.TYPE		Module
.IDENTIFICATION s2d.c
.AUTHOR  	Francois Ochsenbein [ESO]  
.KEYWORDS       Astronomical Coordinate Transformations
.VERSION 1.0    15-Nov-1985: Creation
.VERSION 1.1    26-May-1989: Cosmetic changes
.VERSION 1.2    02-Jun-1989: Added dist_ as functions (not macros)
.VERSION 2.0	22-Feb-1991: Tested on PCs with TuboC
.VERSION 2.1	13-Oct-1998: Simplification of dist_p
.LANGUAGE	C
.COMMENTS
\begin{TeX}

The routines provided in this module all return 
\quad $\sin^2(\frac{d}{2})$, where $d$ is the distance
between 2 points on the celestial sphere.
The position of the points can be specified
as angles (in degrees, {\em s2d\_o}), 
vectors (direction cosines, {\em s2d\_u}), 
or gnomonic projections ({\em s2d\_p}).

The actual distance (in degrees) can be derived with
{\tt 2.e0 * asind(sqrt({\em s2d}))}

On a gnomonic projection, the distance between two points
can be computed by:
$$ \tan(\rho) = \frac{\sqrt{(x_A-x_B)^2 + (y_A-y_B)^2 _ (x_Ay_B-y_Ax_B)^2}}
	{1+x_Ax_B+y_Ay_B} $$
\end{TeX}

--------------------------------------*/
 
#include <trigo.h>

                               /* the following #define help to understand  */
#define ra           q[0] 
#define dec          q[1]
#define ra2          q2[0] 
#define dec2         q2[1]

#define x            u[0]
#define y            u[1]
#define z            u[2]
#define x2           u2[0]
#define y2           u2[1]
#define z2           u2[2]

#define X            p[0]
#define Y            p[1]
#define X2           p2[0]
#define Y2           p2[1]

/*========================================================================*/
double s2d_p(p,p2)   
/*++++++++++++++++++++++
.PURPOSE Compute sin**2(distance/2) from projections on a tangent plane
.Returns sin**2(dist/2)  
.METHOD  
\begin{TeX}
Use relation $(\vec{u_2} - \vec{u})^2 = 2(1 - \cos d) $
\end{TeX}
-----------------------*/
  double p[2]; 	/* IN: projection (X,Y) of the first point */
  double p2[2]; 	/* IN: projection (X,Y) of the second point*/
{
  double f,f2, returned_value, dx; 
  int 	i;
    f  = 1.e0/sqrt(1.e0 + X *X  + Y *Y );
    f2 = 1.e0/sqrt(1.e0 + X2*X2 + Y2*Y2);
    returned_value = f2-f; returned_value *= returned_value;
    for (i=0; i<2; i++)
  	dx = f*p[i] - f2*p2[i], returned_value += dx*dx;
    return (.25e0 * returned_value);
}

/*========================================================================*/
double s2d_o(q, q2)  
/*++++++++++++++++++++++
.PURPOSE Computations of sin**2(dist/2) from equatorial position
.Returns sin**2(dist/2)  
.METHOD  Direct computation
-----------------------*/
  double q[2];	/* IN: position (ra,dec) of the first point */
  double q2[2];	/* IN: position (ra,dec) of the second point*/
{ 
  double sda, sdd;
    sda = sind (.5e0*( ra2 - ra )); sdd = sind (.5e0*(dec2 - dec));
    return ( sdd*sdd + sda*sda * cosd(dec) * cosd(dec2));
}

/*========================================================================*/

double s2d_u(u,u2)  
/*++++++++++++++++++++++
.PURPOSE Compute sin**2(dist/2) from dir. cosines
.Returns sin**2(dist/2)  
.METHOD  
\begin{TeX}
Use relation $(\vec{u_2} - \vec{u})^2 = 2(1 - \cos d) $
\end{TeX}
-----------------------*/
  double u[3];	/* IN: : dir. cosines (x,y,z) of the first point */
  double u2[3];	/* IN: : dir. cosines (x,y,z) of the second point*/
{
  int i;
  double returned_value, dx;
    returned_value = 0.e0;
    for (i=0; i<3; i++)
	dx = u2[i] - u[i], returned_value += dx*dx;
    return (.25e0 * returned_value);
}

/*========================================================================*/
/*		Computation of Distances				  */
/*========================================================================*/

double dist_o(o1, o2)  
/*++++++++++++++++++++++
.PURPOSE Compute distance between two points 
.RETURNS Distance (degrees, range 0 to 180 degrees)
-----------------------*/
  double o1[2];	/* IN: position (ra,dec) of the first point */
  double o2[2];	/* IN: position (ra,dec) of the second point*/
{
    return(2.e0*asind(sqrt(s2d_o(o1, o2))));
}

double dist_p(p0, p1)  
/*++++++++++++++++++++++
.PURPOSE Compute distance between two points 
.RETURNS Distance (degrees, range 0 to 180 degrees)
-----------------------*/
  double p0[2]; 	/* IN: projection (X,Y) of the first point */
  double p1[2]; 	/* IN: projection (X,Y) of the second point*/
{
  double  d, tgr ;
    /* Compute from tan(d) = sqrt(dx**2 + dy**2 + (x1y2-x2y1)**2)/(1+xx+yy) */
    d = p1[0] - p0[0] ; tgr  = d*d ;
    d = p1[1] - p0[1] ; tgr += d*d ;
    d = p1[1]*p0[0] - p0[1]*p1[0]; tgr += d*d ;
    tgr = sqrt(tgr) / (1. + p0[0]*p1[0] + p0[1]*p1[1]) ;
    return(atand(tgr)) ;
    /* return(2.e0*asind(sqrt(s2d_p(p1, p2)))); */
}

double dist_u(u1, u2)  
/*++++++++++++++++++++++
.PURPOSE Compute distance between two points 
.RETURNS Distance (degrees, range 0 to 180 degrees)
-----------------------*/
  double u1[3];	/* IN: : dir. cosines (x,y,z) of the first point */
  double u2[3];	/* IN: : dir. cosines (x,y,z) of the second point*/
{
    return(2.e0*asind(sqrt(s2d_u(u1, u2))));
}

