/*++++++++++
.TYPE           Module
.IDENTIFICATION trmt.c
.AUTHOR         Francois Ochsenbein [ESO-IPG]
.VERSION 1.0	11-May-1990: Creation
.VERSION 1.1	08-Jun-1990: Added more possibilities.
.VERSION 1.2	13-Nov-1992: Added integer cast for shift operations
.VERSION 1.3	27-Nov-1992: Be sure 2nd type is of Hubble type
.VERSION 1.4	25-Oct-1995: Verify the returned error codes
.VERSION 1.5	24-Nov-1995: Allow differentiation between
.VERSION 1.6	08-Feb-2010: 64-bit
				SBb/Sc and   Sb/SBc
.LANGUAGE       C
.KEYWORDS       Morphological Classification

.ENV            Any

.COMMENTS

\begin{TeX}
The Morphological Classification is defined, either as a numeric value
(de Vaucouleurs), or as a Hubble class.
The correspondance is:

$$\begin{tabular}{|r l|}\hline
-6&   Compact Elliptical (cE) \\
-5&   E, ~ dwarf elliptical (dE) \\
-4&   E \\
-3&   L-, S0- \\
-2&   L, S0 \\
-1&   L+, S0+ \\
0&   S0a, S0-a \\
1&   Sa \\
2&   Sab \\
3&   Sb \\
4&   Sbc \\
5&   Sc \\
6&   Scd \\
7&   Sd \\
8&   Sdm \\
9&   Sm, Magellanic Spiral \\
10&   Im, Irr I, Magellanic Irregular, Dwarf Irregular \\
11&   Compact Irregular, Extragalactic HII Region \\
\hline\end{tabular}$$

The pecularities may be {\em B} (barred), {\em p} (peculiar),
{\em r} (ring), {\em D} (double),
{\em s} (?)

The morphological classification is coded as an unsigned integer.
It includes the following bits, from left to right:
$$\begin{tabular}{| l| r r|}\hline Meaning & Length & Mask \\ \hline
    Morphological class (-128 to 127) &	1 byte & 0xff000000 \\
2nd morphological class, only as Hubble class &	1 byte & 0x00ff0000 \\ \hline
Existence of a Bar, with uncertaintly flag & 2 bits & 0x0000C000 \\
Uncertainty flag &	1 bit& 0x00002000 \\
{\bf +} preceding the 2nd morphological type & 1 bit &0x00001000 \\ \hline
Which system (0/1 = numeric with 0/1 decimal, 2 = Hubble) & 2 bits & 0x00000C00 \\
 Unused     				&	1 bit& 0x00000200 \\
Incompleteness flag (translated with \dots) &	1 bit& 0x00000100 \\
2nd Existence of a Bar, with uncertaintly flag & 2 bits & 0x000000C0 \\
Uncertainty flag on 2nd type &	1 bit & 0x0000020 \\
{bf p} flag &	1 bit & 0x00000010 \\
{bf r} flag &	1 bit & 0x00000008 \\
{bf s} flag &	1 bit & 0x00000004 \\
{bf D} flag &	1 bit & 0x00000002 \\
\hline \end{tabular}$$

\end{TeX}

-----------------------------------------------------------------------*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#define MIN(a,b)	((a) < (b) ? a : b )

#define MAX_EDITED	24	/* Maximal length of edited morphological */
#define MT_miss		0x100
#define MT_u1  		0x2000
#define MT_u2  		0x20
#define MT_p		0x10
#define MT_r		0x08
#define MT_s		0x04
#define MT_D		0x02

#ifndef int4
#define int4 int
#endif

#define isSpiral(Gal)	(Gal >= -20) && (Gal <= 90)

static char sspiral[] = { 	/* Sequence of Spiral Galaxies	*/
	'0', 'a', 'b', 'c', 'd', 'm', 0
    };
static char *p, *e, c, format, barred, incomplete, colon;


/*===========================================================================
	Static variables for communications of internal routines
 *===========================================================================*/

static	char	err_level     = 1;	/* 0 = Warning	*/
static	char	display_error = 1;	/* Could be changed */

static struct {				/* Error report	*/
  char *str;		/* Text where error occured	*/
  char *end;		/* Indicates end of str		*/
  int  error;		/* Error level			*/
 } Str;			/* Error Structure 		*/

/*===========================================================================
		Error report
 *===========================================================================*/

static int error(msg, err)
/*++++++
.PURPOSE Display Error.
.RETURNS 0 (no error) / 1 (error)
--------*/
  char	*msg;	/* IN: Error Message */
  char	*err;	/* IN: Position of bad character */
{
  char *p, c;
    if (! msg)	return(0);
    Str.error += err_level;
    if (display_error) {
    	c = *(Str.end), *(Str.end) = 0;
  	fprintf(stderr, "%s%s in: %s\n         ",
  	    err_level ? "****" : "++++", msg, Str.str);
  	for (p = msg; *p; p++) fputc(' ', stderr);
  	for (p = (char *)Str.str; p < err; p++) fputc(' ', stderr);
  	fputc('^', stderr); fputc('\n', stderr);
    	*(Str.end) = c;
    }
    return(1);
}
/*===========================================================================*/
static int NextChar()
/*++++++
.PURPOSE Get next char (skip blanks), and translate special symbols () :
.RETURNS The char, 0 when NULL
.REMARKS I (Irregular) may be written I or Irr
--------*/
{
    while (p < e) {
  	c = *(p++);
  	if (isspace(c))	continue;
  	switch (c) {
  	    case ',':	continue;
  	    case '(':	continue;
  	    case ')':	case ':':
  	    		colon = c;	continue;
	    case 'I':	if ((p[0] == 'r') && (p[1] == 'r')) p += 2;
	    		break;
  	    case '.':	if ((p[0] == '.') && (p[1] == '.')){
  	    			incomplete = 1, p += 2;
  	    			continue;
			}
	    		break;
  	}
  	return(c);
    }
    c = 0;
    return(0);
}

/*=======================================================================*/
static int match(cc)
/*++++++
.PURPOSE Reads until string differs (case insensitive)
.RETURNS Number of matched chars
.REMARKS Match only from second char (it's assumed that c = cc[0])
--------*/
  char *cc;	/* IN: Comparison String */
{
  char 	*q, u;
  int	i, n;

    for (q = cc+1, n=0; p[n]; n++)  {
  	u = tolower(p[n]);
  	if (u != q[n])	break;
    }
    for (i = n; --i >= 0; )	NextChar();
    return(n);
}

/*===========================================================================*/
static int get_hubble()
/*++++++
.PURPOSE Converts the string to a Hubble class.
.RETURNS The Morphological class (integer in range -128 / 127)
.REMARKS If an error is encountered, error displayed
--------*/
{
  int	mt, i, old_colon;

    format = 2;			/* It'a s Hubble type	*/
    colon = barred = 0;
    mt = -128;				/*     Default 	*/

    switch(c) {
    case '?': 	NextChar();	/* NO break	*/
    case 0:
    			break;
    case 'c':	case 'C':	/* Compact */
    	match("compact");
	mt = -60;	goto dwarf;
    case 'd':  	case 'D':	/* Dwarf (or double?)  */
	if ((*p == 'o') || (*p == 'b'))	break;
	mt = -50;
    dwarf :			/* Followed by E / G / Irr */
	match("dwarf");
	NextChar();
	switch(c) {
	    case 'E': 	NextChar(); 	break;
	    case 'G':	NextChar();
	    default:	mt -= 5;	break;
	    case 'I':	NextChar();
	    		mt = 80 - mt/2;	break;
	}
	break;
    case 'E':	/*Elliptical	*/
	mt = -40;
	NextChar(); 	break;
    case 'L':	/* Lenticular	*/
	mt = -20;
	NextChar(); 	break;
    case 'S':	/* Spiral	*/
	NextChar();
	if (c == 'B')	{	/* Barred Spiral	*/
	    old_colon = colon;	colon = 0;
	    barred = 1;	NextChar();
	    if (colon) 	barred = 2;
	    colon = old_colon;
	}
	for (i=0; sspiral[i] && (sspiral[i] != c); i++) ;
	/* i = strloc (sspiral, c); */
	if (sspiral[i]) {	/* Standard Spiral 	*/
	    NextChar();	mt = i*20 - 10;
	    if ((c) && (sspiral[i+1] == c)) {	/* eg Sab */
		NextChar();
		mt += 10;
	    }
	    if (mt < 0)	mt = -20;	/* Only for S0	*/
	}
	else mt = 45;			/* S alone ...	*/
	break;
    case 'I':	/* Irregular	*/
	NextChar();	mt = 100;
	if (c == 'm') { NextChar(); mt = 101; }
	break;
    default:	/* May be a pecularity	*/
	/* error("Invalid Hubble Type", p); */
    	break;
    }
#ifdef TEST
    printf("....get_hubble returns %d\n", mt);
#endif
    return(mt);
}

/*===========================================================================*/
static int get_dV()
/*++++++
.PURPOSE Converts the string to deVaucouleurs... + Bar.
.RETURNS The Class
.REMARKS If an error is encountered, error written
--------*/
{
  int	sign, mt, old_colon;

    mt = 		/* Default */
    sign = format = colon = barred = 0;

    if (c == '-')	{ NextChar(); sign = 1; }
    while (isdigit(c)) { mt = mt*10 + (c-'0'); NextChar(); }

    mt *= 10;
    if ((c == '.') && (*p != '.')) {	/* There is a decimal */
	NextChar();
	format = 1;
	mt += (c - '0');
	NextChar();
    }
    if (sign)	mt = -mt;

    if (c == 'B')	{	/* Barred Spiral	*/
	old_colon = colon;	colon = 0;
	barred = 1;	NextChar();
	if (colon) 	barred = 2;
	colon = old_colon;
    }

	/* Check that mt is in correct interval	*/

    if (mt <= -70)	error("deVaucouleurs Type too small", p);
    if (mt >  110)	error("deVaucouleurs Type too large", p);

    return(mt);
}

/*===========================================================================*/
int tr_akm(str, len, mt)
/*++++++
.PURPOSE Converts a string containing a morphological class to a integer;
.RETURNS -1 for error, 0 for completely coded string, 1 if incomplete
.REMARKS If an error is encountered, message logged.
--------*/
  char	*str;	/* IN: String to scan	*/
  int	len;	/* IN: Length of str	*/
  int4	*mt;	/* OUT: result		*/
{
  int4	m1, m2;
  int	mp, second;

    Str.str = p = str;
    Str.end = e = (str + len);
    Str.error = 0;
    incomplete = 0;
    mp = second = 0;

	/* Check first if deVaucouleurs / Hubble */

    NextChar();
    if (isdigit(c))		m1 = get_dV();
    else if (isalpha(c))	m1 = get_hubble();
    else if (isdigit(*p))	m1 = get_dV();
    else			m1 = get_hubble();

				/* Undefined type is universal.. */
    if (m1 == -128)		format = 0;

	/* Save Pecularities	*/

    mp = ((int)barred << 14) | ((int)format << 10);
    if (colon)	mp |= MT_u1;
    barred = colon = 0;

	/* Get Second Type, if any	*/

    m2 = -128; second = 0;
    if (c == '+')	{ mp |= 0x1000; second = 1; NextChar(); }
    else if (c == '/')	{ second = 1; NextChar(); }
    if (second && (format == 2)) {	/* Only for Hubble type */
	if (colon)	mp |= MT_u2;
	m2 = get_hubble();
        /* m2 = (format == 2 ? get_hubble() : get_dV()); */
        if (barred) mp |= ((int)barred << 6);
	if (colon)	mp |= MT_u2;
    }

	/* Pecularities */

    while(c) {
	switch(tolower(c)) {
	  case 'p':	mp |= MT_p;	match("peculiar");	break;
	  case 'r':	mp |= MT_r;				break;
	  case 's':	mp |= MT_s;				break;
	  case 'd':	mp |= MT_D;
	  		if (*p == 'b')	match("dble");
	  		else		match("double");	break;
	  default :	mp |= MT_miss;				break;
	}
	NextChar();
    }

    if (incomplete)	mp |= MT_miss;
    *mt = (m1 << 24L) | ((m2&0xff)<<16L) | (mp&0xffff);
    if (Str.error)
  	 mp = -1;
    else mp = (mp & MT_miss ? 1 : 0);

    return(mp)  ;
}

/*===========================================================================*/
static char *ed_dV(mt, nd)
/*++++++
.PURPOSE Edit de Vaucouleurs morphological type
.RETURNS Pointer to edited string
.REMARKS
--------*/
  int	mt;	/* IN: Morphological type (-128 / 127)	*/
  int	nd;	/* IN: Number of decimals (0 or 1)	*/
{
  static	char buf[6];
  char	*p;

    p  = buf;
    *p = 0;
    if (mt <= -128)	return(p);

    if (mt >= 100)	*p = '1';
    else if (mt < 0)	*p = '-', mt = -mt;
    else		*p = ' ';
    p++;
    mt %= 100;
    *(p++) = '0' + mt/10;
    if (nd) {
	*(p++) = '.';
	*(p++) = '0' + mt % 10;
    }
    if (barred)  {
	*(p++) = 'B';
	if (barred & 2) *(p++) = ':';
    }
    *p = 0;
    return (buf);
}

/*===========================================================================*/
static char *ed_hubble(mt)
/*++++++
.PURPOSE Edit Hubble morphological type
.RETURNS Length of edited string
.REMARKS
--------*/
  int	mt;	/* IN: Morphological type (-128 / 127)	*/
{
  char	*p;
  static char buf[6];
  int	i, k, bar;

    p = buf; 	*p = 0;
    if (mt <= -128)	return(p);
    bar = barred;

    if (isSpiral(mt))   {	/* Spiral Galaxies */
	i = (mt < 0 ? 0 : mt+10);
	*(p++) = 'S';
	if (bar)  {
	    *(p++) = 'B';
	    if (bar & 2) *(p++) = ':';
	    bar = 0;
	}
	if (i != 55) {
	    k = i/20;
	    *(p++) = sspiral[k];
	    switch (i%20) {
		case 0:					break;
		case 10: *(p++) = sspiral[k+1]; 	break;
		default: *(p++) = '?';			break;
	    }
	}
    } else if (mt == -40)	*(p++) = 'E';
    else   if (mt <= -50) {
	k = -50 - mt;
	*(p++) = (k/10 ? 'c' : 'd');
	*(p++) = (k%10 ? 'G' : 'E');
	if ((k%5) || (k > 15))	*(p++) = '?';
	*p = 0;
    }
    else if (mt == 100)	*(p++) = 'I';
    else if (mt == 101)	*(p++) = 'I', *(p++) = 'm';
    else if (mt == 105)	*(p++) = 'd', *(p++) = 'I';
    else if (mt == 110)	*(p++) = 'c', *(p++) = 'I';
    else		*(p++) = '?';
    if (bar)  {
	*(p++) = 'B';
	if (bar & 2) *(p++) = ':';
    }
    *p = 0;

    return(buf);
}

/*===========================================================================*/
char *kmtoa(mt)
/*++++++
.PURPOSE Edit a Coded Morphological class
.RETURNS Address of edited Spectral Class (terminated with an 0)
.REMARKS
--------*/
  int4	mt;	/* IN: Morphological Class */
{
  int	m;
  char	*p, *q;
  static  char	mted[MAX_EDITED];

    format = (mt >> 10) & 3;
    barred = (mt >> 14) & 3;
    p = mted;
    if (mt == 0x80800000)	return("");

    m = mt >> 24;
    if   (m == -128) 		q = "?";
    else if (format < 2)	q = ed_dV (m, format);
    else 			q = ed_hubble(m);
    strcpy(p, q); p += strlen(p);
    if (mt & MT_u1)	*(p++) = ':';

    format = 0;
    barred = (mt >> 6) & 3;
    m = (mt << 8) >> 24;
    if (mt & 0x1000)	*(p++) = '+';
    else if (m != -128)	*(p++) = '/';
    if (m != -128)	strcpy(p, ed_hubble(m)), p += strlen(p);
    if (mt & MT_u2)	*(p++) = ':';

	/* Now, edit Pecularities */

    if (mt & MT_r)	*(p++) = 'r';
    if (mt & MT_s)	*(p++) = 's';
    if (mt & MT_p)	*(p++) = 'p';
    if (mt & MT_D)	*(p++) = 'D';
    if (mt & MT_miss)	*(p++) = '.', *(p++) = '.', *(p++) = '.';

    *p = 0;

    return(mted);
}

/*===========================================================================*/
int ed_km(str, size, mt)
/*++++++
.PURPOSE Edit a Coded Morphological Type
.RETURNS Length of edited string (may be up to 36 bytes)
.REMARKS If str is too small, edition is truncated. But the returned
	length is still the required length.
--------*/
  char	*str;	/* OUT: Output string	*/
  int	size;	/* IN: Size of str	*/
  int4	mt;	/* IN: Morphological Class */
{
  int	l;
  char	*edited;

    edited = kmtoa(mt);		/* Edit in local buffer	*/
    l = strlen(edited)+ 1;	/* Length of output, including 0	*/
    l = MIN(l, size);
    strncpy(str, edited, l);

    return(l-1);
}

#ifdef TEST  		/* Test Program */

static char msg[132], stred[132];

main()
{
  int4	mt;
  int 	i, n, f0, f1, stat;

    puts("Test of Morphological Classification \n");

    for (;;) {
    	printf("Input Class: ");
    	if (!gets (msg)) 	break;
    	stat = tr_akm(msg, strlen(msg), &mt);
    	printf("--->%08lX %ld%c%ld  status=%d\n", mt, mt>>24,
		(mt & 0x1000 ? '+':'/'), (mt<<8)>>24, stat);
    	puts(kmtoa(mt));
    }
    exit(0);
}
#endif
